'use strict';

const buffer = require('buffer');
const EventEmitter = require('eventemitter3');
const url = require('url');
const http = require('http');
const https = require('https');
const stream$2 = require('stream');
const followRedirects = require('follow-redirects');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e["default"] : e; }

const EventEmitter__default = /*#__PURE__*/_interopDefaultLegacy(EventEmitter);
const http__default = /*#__PURE__*/_interopDefaultLegacy(http);
const https__default = /*#__PURE__*/_interopDefaultLegacy(https);
const followRedirects__default = /*#__PURE__*/_interopDefaultLegacy(followRedirects);

const redirectRegex = /^201|30(1|2|7|8)$/;
function removeChunked(req, res, proxyRes) {
  if (req.httpVersion === "1.0" || proxyRes.statusCode === 204 || proxyRes.statusCode === 304) {
    delete proxyRes.headers["transfer-encoding"];
  }
}
function setConnection(req, res, proxyRes) {
  if (req.httpVersion === "1.0") {
    proxyRes.headers.connection = req.headers.connection || "close";
  } else if (req.httpVersion !== "2.0" && !proxyRes.headers.connection) {
    proxyRes.headers.connection = req.headers.connection || "keep-alive";
  }
}
function setRedirectHostRewrite(req, res, proxyRes, options) {
  if ((options.hostRewrite || options.autoRewrite || options.protocolRewrite) && proxyRes.headers["location"] && redirectRegex.test(proxyRes.statusCode)) {
    const targetStr = typeof options.target === "string" ? options.target : options.target.href;
    let target = new URL(targetStr);
    const u = new URL(proxyRes.headers["location"]);
    if (target.host != u.host) {
      return;
    }
    if (options.hostRewrite) {
      u.host = options.hostRewrite;
    } else if (options.autoRewrite) {
      u.host = req.headers["host"];
    }
    if (options.protocolRewrite) {
      u.protocol = options.protocolRewrite;
    }
    proxyRes.headers["location"] = u.toString();
  }
}
function writeHeaders(req, res, proxyRes, options) {
  const preserveHeaderKeyCase = options.preserveHeaderKeyCase;
  let rewriteCookieDomainConfig = options.cookieDomainRewrite, rewriteCookiePathConfig = options.cookiePathRewrite, mergeCookiesConfig = options.mergeCookies, cookieRemoveSecure = options.cookieRemoveSecure, rawHeaderKeyMap;
  const setHeader = function(key, header) {
    if (rewriteCookieDomainConfig && key.toLowerCase() === "set-cookie") {
      header = rewriteCookieProperty(header, rewriteCookieDomainConfig, "domain");
    }
    if (rewriteCookiePathConfig && key.toLowerCase() === "set-cookie") {
      header = rewriteCookieProperty(header, rewriteCookiePathConfig, "path");
    }
    if (mergeCookiesConfig && key.toLowerCase() === "set-cookie") {
      header = mergeSetCookie(res.getHeader("set-cookie"), header);
    }
    if (cookieRemoveSecure && key.toLowerCase() === "set-cookie") {
      header = removeCookieProperty(header, "secure");
    }
    try {
      res.setHeader(String(key).trim(), header);
    } catch (error) {
      console.warn(error, key, header);
    }
  };
  if (typeof rewriteCookieDomainConfig === "string") {
    rewriteCookieDomainConfig = { "*": rewriteCookieDomainConfig };
  }
  if (typeof rewriteCookiePathConfig === "string") {
    rewriteCookiePathConfig = { "*": rewriteCookiePathConfig };
  }
  if (preserveHeaderKeyCase && proxyRes.rawHeaders != void 0) {
    rawHeaderKeyMap = {};
    for (let i = 0; i < proxyRes.rawHeaders.length; i += 2) {
      let key = proxyRes.rawHeaders[i];
      rawHeaderKeyMap[key.toLowerCase()] = key;
    }
  }
  Object.keys(proxyRes.headers).forEach(function(key) {
    const header = proxyRes.headers[key];
    if (preserveHeaderKeyCase && rawHeaderKeyMap) {
      key = rawHeaderKeyMap[key] || key;
    }
    setHeader(key, header);
  });
}
function attachOutgoingHeaders(req, res, proxyRes, options) {
  if (options.outgoingHeaders != null) {
    Object.keys(options.outgoingHeaders).forEach(function(header) {
      res.setHeader(header, options.outgoingHeaders[header]);
    });
  }
}
function writeStatusCode(req, res, proxyRes) {
  if (proxyRes.statusMessage) {
    res.statusCode = proxyRes.statusCode;
    res.statusMessage = proxyRes.statusMessage;
  } else {
    res.statusCode = proxyRes.statusCode;
  }
}
const webOutgoing = {
  removeChunked,
  setConnection,
  setRedirectHostRewrite,
  writeHeaders,
  attachOutgoingHeaders,
  writeStatusCode
};

const passes = Object.keys(webOutgoing).map(function(pass) {
  return webOutgoing[pass];
});
const isSSL = /^https|wss/;
const upgradeHeader = /(^|,)\s*upgrade\s*($|,)/i;
function required(port, protocol) {
  protocol = protocol.split(":")[0];
  port = +port;
  if (!port)
    return false;
  switch (protocol) {
    case "http":
    case "ws":
      return port !== 80;
    case "https":
    case "wss":
      return port !== 443;
    case "ftp":
      return port !== 21;
    case "gopher":
      return port !== 70;
    case "file":
      return false;
  }
  return port !== 0;
}
function setupOutgoing(outgoing, options, req, forward) {
  const target = options[forward || "target"];
  if (typeof target === "object") {
    if (!target.searchParams) {
      target.searchParams = new URLSearchParams();
    }
    if (target.path) {
      target.pathname = target.path;
      delete target.path;
    }
  }
  const sslEnabled = isSSL.test(target.protocol);
  outgoing.port = target.port || (sslEnabled ? 443 : 80);
  ["host", "hostname", "socketPath", "pfx", "key", "passphrase", "cert", "ca", "ciphers", "secureProtocol", "servername"].forEach(function(e) {
    outgoing[e] = target[e];
  });
  outgoing.method = options.method || req.method;
  outgoing.headers = Object.assign({}, req.headers);
  if (options.headers) {
    Object.assign(outgoing.headers, options.headers);
  }
  if (options.auth) {
    outgoing.auth = options.auth;
  }
  if (options.ca) {
    outgoing.ca = options.ca;
  }
  if (sslEnabled) {
    outgoing.rejectUnauthorized = typeof options.secure === "undefined" ? true : options.secure;
  }
  outgoing.agent = options.agent || false;
  outgoing.localAddress = options.localAddress;
  if (!outgoing.agent) {
    outgoing.headers = outgoing.headers || {};
    if (typeof outgoing.headers.connection !== "string" || !upgradeHeader.test(outgoing.headers.connection)) {
      outgoing.headers.connection = "close";
    }
  }
  if (options.followRedirects && typeof options.followRedirects === "object") {
    outgoing.maxRedirects = options.followRedirects.maxRedirects;
    outgoing.maxBodyLength = options.followRedirects.maxBodyLength;
    outgoing.agents = options.followRedirects.agents;
    outgoing.beforeRedirect = options.followRedirects.beforeRedirect;
    outgoing.trackRedirects = options.followRedirects.trackRedirects;
  }
  const targetPath = target && options.prependPath !== false ? target.pathname || "" : "";
  const reqUrl = new URL(req.url, "http://example.com");
  for (const entry of target.searchParams.entries()) {
    reqUrl.searchParams.set(entry[0], entry[1]);
  }
  const params = reqUrl.search;
  let outgoingPath = !options.toProxy ? reqUrl.pathname || "" : req.url;
  outgoingPath = !options.ignorePath ? outgoingPath : "";
  outgoing.path = [targetPath, outgoingPath].filter(Boolean).join("/").replace(/\/+/g, "/") + params;
  if (options.changeOrigin) {
    outgoing.headers.host = required(outgoing.port, target.protocol) && !hasPort(outgoing.host) ? outgoing.host + ":" + outgoing.port : outgoing.host;
  }
  return outgoing;
}
function setupSocket(socket) {
  socket.setTimeout(0);
  socket.setNoDelay(true);
  socket.setKeepAlive(true, 0);
  return socket;
}
function getPort(req) {
  const res = req.headers.host ? req.headers.host.match(/:(\d+)/) : "";
  return res ? res[1] : hasEncryptedConnection(req) ? "443" : "80";
}
function hasEncryptedConnection(req) {
  return Boolean(req.connection.encrypted || req.connection.pair);
}
function rewriteCookieProperty(header, config, property) {
  if (Array.isArray(header)) {
    return header.map(function(headerElement) {
      return rewriteCookieProperty(headerElement, config, property);
    });
  }
  return header.replace(new RegExp("(;\\s*" + property + "=)([^;]+)", "i"), function(match, prefix, previousValue) {
    let newValue;
    if (previousValue in config) {
      newValue = config[previousValue];
    } else if ("*" in config) {
      newValue = config["*"];
    } else {
      return match;
    }
    if (newValue) {
      return prefix + newValue;
    } else {
      return "";
    }
  });
}
function removeCookieProperty(header, property) {
  if (Array.isArray(header)) {
    return header.map(function(headerElement) {
      return removeCookieProperty(headerElement, property);
    });
  }
  return header.replace(new RegExp(";\\s*" + property + "[^;]*", "i"), "");
}
function mergeSetCookie(setCookie, upstreamSetCookie) {
  let existingCookieArray = setCookie || [], upstreamCookieArray = upstreamSetCookie || [];
  if (!Array.isArray(existingCookieArray)) {
    existingCookieArray = [existingCookieArray];
  }
  if (!Array.isArray(upstreamCookieArray)) {
    upstreamCookieArray = [upstreamCookieArray];
  }
  return [].concat(existingCookieArray, upstreamCookieArray);
}
function runWebOutgoingPasses(req, res, proxyRes, options) {
  for (let i = 0; i < passes.length; i++) {
    console.log(passes);
    if (passes[i](req, res, proxyRes, options)) {
      break;
    }
  }
}
function hasPort(host) {
  return !!~host.indexOf(":");
}

const nativeAgents = { http: http__default, https: http__default };
const hasAbortedEvent = function() {
  var ver = process.versions.node.split(".").map(Number);
  return ver[0] <= 14 || ver[0] === 15 && ver[1] <= 4;
}();
function deleteLength(req, res, options) {
  if ((req.method === "DELETE" || req.method === "OPTIONS") && !req.headers["content-length"]) {
    req.headers["content-length"] = "0";
    delete req.headers["transfer-encoding"];
  }
}
function timeout(req, res, options) {
  let timeoutValue = options.timeout ? options.timeout : 0;
  req.socket.setTimeout(timeoutValue);
}
function XHeaders$1(req, res, options) {
  if (!options.xfwd)
    return;
  let encrypted = req.isSpdy || hasEncryptedConnection(req);
  let values = {
    for: req.connection.remoteAddress || req.socket.remoteAddress,
    port: getPort(req),
    proto: encrypted ? "https" : "http"
  };
  ["for", "port", "proto"].forEach(function(header) {
    req.headers["x-forwarded-" + header] = (req.headers["x-forwarded-" + header] || "") + (req.headers["x-forwarded-" + header] ? "," : "") + values[header];
  });
  req.headers["x-forwarded-host"] = req.headers["x-forwarded-host"] || req.headers["host"] || "";
}
function stream$1(req, res, options, _, server, callback) {
  server.emit("start", req, res, options.target || options.forward);
  const agents = options.followRedirects ? followRedirects__default : nativeAgents;
  const http2 = agents.http;
  const https2 = agents.https;
  if (options.forward) {
    const forwardReq = (isSSL.test(options.forward.protocol) ? https2 : http2).request(setupOutgoing(options.ssl || {}, options, req, "forward"));
    const forwardError = createErrorHandler(forwardReq, options.forward);
    req.on("error", forwardError);
    forwardReq.on("error", forwardError);
    stream$2.pipeline(options.buffer || req, forwardReq, () => {
    });
    if (!options.target) {
      return res.end();
    }
  }
  const proxyReq = (isSSL.test(options.target.protocol) ? https2 : http2).request(setupOutgoing(options.ssl || {}, options, req));
  proxyReq.on("socket", function(socket) {
    if (server && !proxyReq.getHeader("expect")) {
      server.emit("proxyReq", proxyReq, req, res, options);
    }
  });
  if (options.proxyTimeout) {
    proxyReq.setTimeout(options.proxyTimeout, function() {
      proxyReq.abort();
    });
  }
  if (hasAbortedEvent) {
    req.on("aborted", function() {
      proxyReq.abort();
    });
  } else {
    res.on("close", function() {
      var aborted = !res.writableFinished;
      if (aborted) {
        proxyReq.abort();
      }
    });
  }
  const proxyError = createErrorHandler(proxyReq, options.target);
  proxyReq.on("error", proxyError);
  req.on("error", proxyError);
  function createErrorHandler(proxyReq2, url) {
    return function proxyError2(err) {
      if ((req.aborted || req.socket.destroyed) && err.code === "ECONNRESET") {
        server.emit("econnreset", err, req, res, url);
        return proxyReq2.abort();
      }
      if (callback) {
        callback(err, req, res, url);
      } else {
        server.emit("error", err, req, res, url);
      }
    };
  }
  stream$2.pipeline(options.buffer || req, proxyReq, () => {
  });
  proxyReq.on("response", function(proxyRes) {
    proxyRes.abort = function() {
      proxyRes.aborted = true;
      proxyReq.abort();
      proxyRes.abort = function() {
      };
    };
    if (server) {
      server.emit("proxyRes", proxyRes, req, res, options);
    }
    const selfHandle = typeof options.selfHandleResponse === "function" ? options.selfHandleResponse(proxyRes, req, res) : options.selfHandleResponse;
    if (!proxyRes.aborted) {
      if (!res.headersSent && (!selfHandle || options.forcePasses)) {
        runWebOutgoingPasses(req, res, proxyRes, options);
      }
      if (!res.finished) {
        proxyRes.on("end", function() {
          if (server) {
            server.emit("end", req, res, proxyRes);
          }
        });
        if (!selfHandle) {
          stream$2.pipeline(proxyRes, res, () => {
          });
        }
      } else {
        if (server) {
          server.emit("end", req, res, proxyRes);
        }
      }
    }
  });
}
const webIncoming = {
  deleteLength,
  timeout,
  XHeaders: XHeaders$1,
  stream: stream$1
};

function checkMethodAndHeader(req, socket) {
  if (req.method !== "GET" || !req.headers.upgrade) {
    socket.destroy();
    return true;
  }
  if (req.headers.upgrade.toLowerCase() !== "websocket") {
    socket.destroy();
    return true;
  }
}
function XHeaders(req, socket, options) {
  if (!options.xfwd)
    return;
  let values = {
    for: req.connection.remoteAddress || req.socket.remoteAddress,
    port: getPort(req),
    proto: hasEncryptedConnection(req) ? "wss" : "ws"
  };
  ["for", "port", "proto"].forEach(function(header) {
    req.headers["x-forwarded-" + header] = (req.headers["x-forwarded-" + header] || "") + (req.headers["x-forwarded-" + header] ? "," : "") + values[header];
  });
}
async function stream(req, socket, options, head, server, callback) {
  let createHttpHeader = function(line, headers) {
    return Object.keys(headers).reduce(function(head2, key) {
      let value = headers[key];
      if (!Array.isArray(value)) {
        head2.push(key + ": " + value);
        return head2;
      }
      for (let i = 0; i < value.length; i++) {
        head2.push(key + ": " + value[i]);
      }
      return head2;
    }, [line]).join("\r\n") + "\r\n\r\n";
  };
  setupSocket(socket);
  if (head && head.length) {
    socket.unshift(head);
  }
  let proxyReq = (isSSL.test(options.target.protocol) ? https__default : http__default).request(setupOutgoing(options.ssl || {}, options, req));
  proxyReq.on("error", onOutgoingError);
  proxyReq.on("response", function(res) {
    if (!res.upgrade && !socket.destroyed) {
      socket.write(createHttpHeader("HTTP/" + res.httpVersion + " " + res.statusCode + " " + res.statusMessage, res.headers));
      res.pipe(socket).on("error", onOutgoingError);
    }
  });
  proxyReq.on("upgrade", function(proxyRes, proxySocket, proxyHead) {
    proxySocket.on("error", onOutgoingError);
    proxySocket.on("end", function() {
      server.emit("close", proxyRes, proxySocket, proxyHead);
    });
    socket.on("error", function() {
      proxySocket.end();
    });
    setupSocket(proxySocket);
    if (proxyHead && proxyHead.length) {
      proxySocket.unshift(proxyHead);
    }
    if (!req.headers["x-method-ori"]) {
      socket.write("HTTP/1.1 101 Switching Protocols\r\n");
      socket.write(Object.keys(proxyRes.headers).map(function(i) {
        return i + ": " + proxyRes.headers[i];
      }).join("\r\n") + "\r\n\r\n");
    }
    let proxyStream = proxySocket;
    if (options.createWsServerTransformStream) {
      const wsServerTransformStream = options.createWsServerTransformStream(req, proxyReq, proxyRes);
      wsServerTransformStream.on("error", onOutgoingError);
      proxyStream = proxyStream.pipe(wsServerTransformStream);
    }
    proxyStream = proxyStream.pipe(socket);
    if (options.createWsClientTransformStream) {
      const wsClientTransformStream = options.createWsClientTransformStream(req, proxyReq, proxyRes);
      wsClientTransformStream.on("error", onOutgoingError);
      proxyStream = proxyStream.pipe(wsClientTransformStream);
    }
    proxyStream.pipe(proxySocket);
    server.emit("open", proxySocket);
    server.emit("proxySocket", proxySocket);
  });
  if (server) {
    let asyncHandler;
    const asyncContext = (callback2) => {
      asyncHandler = callback2;
    };
    server.emit("proxyReqWs", proxyReq, req, socket, options, head, asyncContext);
    if (asyncHandler) {
      await asyncHandler();
    }
  }
  proxyReq.end();
  function onOutgoingError(err) {
    if (callback) {
      callback(err, req, socket);
    } else {
      server.emit("error", err, req, socket);
    }
    socket.end();
  }
}
const wsIncoming = {
  checkMethodAndHeader,
  XHeaders,
  stream
};

const webPasses = Object.keys(webIncoming).map(function(pass) {
  return webIncoming[pass];
});
const wsPasses = Object.keys(wsIncoming).map(function(pass) {
  return wsIncoming[pass];
});
class ProxyServer extends EventEmitter__default {
  constructor(options) {
    super();
    options = options || {};
    options.prependPath = options.prependPath === false ? false : true;
    this.web = this.proxyRequest = createRightProxy("web")(options);
    this.ws = this.proxyWebsocketRequest = createRightProxy("ws")(options);
    this.options = options;
    this.webPasses = webPasses;
    this.wsPasses = wsPasses;
    this.on("error", this.onError, this);
  }
  onError(err) {
    if (this.listeners("error").length === 1) {
      throw err;
    }
  }
  listen(port, hostname) {
    const closure = (req, res) => {
      this.web(req, res);
    };
    this.$server = this.options.ssl ? https__default.createServer(this.options.ssl, closure) : http__default.createServer(closure);
    if (this.options.ws) {
      this.$server.on("upgrade", (req, socket, head) => {
        this.ws(req, socket, head);
      });
    }
    this.$server.listen(port, hostname);
    return this;
  }
  close(callback) {
    if (this.$server) {
      this.$server.close(done);
    }
    function done() {
      this.$server = null;
      if (callback) {
        callback.apply(null, arguments);
      }
    }
  }
  before(type, passName, callback) {
    if (type !== "ws" && type !== "web") {
      throw new Error("type must be `web` or `ws`");
    }
    const passes = type === "ws" ? this.wsPasses : this.webPasses;
    let i = false;
    passes.forEach(function(v, idx) {
      if (v.name === passName)
        i = idx;
    });
    if (i === false)
      throw new Error("No such pass");
    passes.splice(i, 0, callback);
  }
  after(type, passName, callback) {
    if (type !== "ws" && type !== "web") {
      throw new Error("type must be `web` or `ws`");
    }
    const passes = type === "ws" ? this.wsPasses : this.webPasses;
    let i = false;
    passes.forEach(function(v, idx) {
      if (v.name === passName)
        i = idx;
    });
    if (i === false)
      throw new Error("No such pass");
    passes.splice(i++, 0, callback);
  }
}
function createRightProxy(type) {
  return function(options) {
    return function(req, res) {
      const passes = type === "ws" ? wsPasses : webPasses;
      let args = [].slice.call(arguments);
      let cntr = args.length - 1;
      let head;
      let callback;
      if (typeof args[cntr] === "function") {
        callback = args[cntr];
        cntr--;
      }
      let requestOptions = options;
      if (!(args[cntr] instanceof buffer.Buffer) && args[cntr] !== res) {
        requestOptions = Object.assign({}, options);
        Object.assign(requestOptions, args[cntr]);
        cntr--;
      }
      if (args[cntr] instanceof buffer.Buffer) {
        head = args[cntr];
      }
      ["target", "forward"].forEach(function(e) {
        if (typeof requestOptions[e] === "string") {
          requestOptions[e] = new url.URL(requestOptions[e]);
        }
      });
      if (!requestOptions.target && !requestOptions.forward) {
        return this.emit("error", new Error("Must provide a proper URL as target"));
      }
      for (let i = 0; i < passes.length; i++) {
        if (passes[i](req, res, requestOptions, head, this, callback)) {
          break;
        }
      }
    };
  };
}

function createProxyServer(options) {
  return new ProxyServer(options);
}
ProxyServer.createProxyServer = createProxyServer;
ProxyServer.createServer = createProxyServer;
ProxyServer.createProxy = createProxyServer;

module.exports = ProxyServer;
